// Copyright 2016 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
// Modified by Jeremy Retailleau.

#include "./nullPtr.h"

namespace pxr {

const TfNullPtrType TfNullPtr = TfNullPtrType();

}  // namespace pxr
